/******************************************************************************\

          This file is part of the C! library.  A.K.A the cbang library.

                Copyright (c) 2003-2019, Cauldron Development LLC
                   Copyright (c) 2003-2017, Stanford University
                               All rights reserved.

         The C! library is free software: you can redistribute it and/or
        modify it under the terms of the GNU Lesser General Public License
       as published by the Free Software Foundation, either version 2.1 of
               the License, or (at your option) any later version.

        The C! library is distributed in the hope that it will be useful,
          but WITHOUT ANY WARRANTY; without even the implied warranty of
        MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
                 Lesser General Public License for more details.

         You should have received a copy of the GNU Lesser General Public
                 License along with the C! library.  If not, see
                         <http://www.gnu.org/licenses/>.

        In addition, BSD licensing may be granted on a case by case basis
        by written permission from at least one of the copyright holders.
           You may request written permission by emailing the authors.

                  For information regarding this software email:
                                 Joseph Coffland
                          joseph@cauldrondevelopment.com

\******************************************************************************/

#include "BufferEvent.h"
#include "Buffer.h"
#include "Base.h"

#include <cbang/Exception.h>
#include <cbang/log/Logger.h>

#include <event2/bufferevent.h>

#ifdef HAVE_OPENSSL
#include <cbang/openssl/SSL.h>
#include <cbang/openssl/SSLContext.h>

#include <event2/bufferevent_ssl.h>

#include <openssl/ssl.h>
#endif // HAVE_OPENSSL

using namespace std;
using namespace cb;
using namespace cb::Event;


BufferEvent::BufferEvent(bufferevent *bev, bool deallocate) :
  bev(bev), deallocate(deallocate) {
}


BufferEvent::BufferEvent(cb::Event::Base &base,
                         const SmartPointer<SSLContext> &sslCtx,
                         const string &host) : bev(0), deallocate(true) {
  if (sslCtx.isNull())
    bev = bufferevent_socket_new(base.getBase(), -1, BEV_OPT_CLOSE_ON_FREE);

#ifdef HAVE_OPENSSL
  else {
    ::SSL *ssl = SSL_new(sslCtx->getCTX());

#ifdef SSL_CTRL_SET_TLSEXT_HOSTNAME
    if (!host.empty() && host.find_first_not_of("1234567890.") != string::npos)
      SSL_set_tlsext_host_name(ssl, host.c_str());
#endif

    bev = bufferevent_openssl_socket_new
      (base.getBase(), -1, ssl, BUFFEREVENT_SSL_CONNECTING,
       BEV_OPT_CLOSE_ON_FREE | BEV_OPT_DEFER_CALLBACKS);
  }
#else
  else THROW("C! was not built with OpenSSL support");

#endif // HAVE_OPENSSL


  if (!bev) THROW("Failed to create buffer event");

#ifdef HAVE_OPENSSL
  bufferevent_openssl_set_allow_dirty_shutdown(bev, 1);
#endif
}


BufferEvent::~BufferEvent() {if (bev && deallocate) bufferevent_free(bev);}


void BufferEvent::setPriority(int priority) {
  bufferevent *ubev = bufferevent_get_underlying(bev);
  if (ubev) BufferEvent(ubev, false).setPriority(priority);
  else if (bufferevent_priority_set(bev, priority))
    THROWS("Unable to set buffer event priority to " << priority);
}


bool BufferEvent::hasSSL() const {
#ifdef HAVE_OPENSSL
  return bufferevent_openssl_get_ssl(bev);
#else
  return false;
#endif
}


#ifdef HAVE_OPENSSL
cb::SSL BufferEvent::getSSL() const {
  struct ssl_st *ssl = bufferevent_openssl_get_ssl(bev);
  if (!ssl) THROW("BufferEvent does not have SSL");
  return cb::SSL(ssl);
}
#endif


void BufferEvent::logSSLErrors() {
#ifdef HAVE_OPENSSL
  string errors = getSSLErrors();
  if (!errors.empty()) LOG_ERROR("SSL errors: " << errors);
#endif
}


string BufferEvent::getSSLErrors() {
  string errors;

#ifdef HAVE_OPENSSL
  unsigned error;

  while ((error = bufferevent_get_openssl_error(bev))) {
    if (!errors.empty()) errors += ", ";
    errors += SSL::getErrorStr(error);
  }
#endif

  return errors;
}


bool BufferEvent::isWrapper() const {return bufferevent_get_underlying(bev);}


BufferEvent BufferEvent::getUnderlying() const {
  return BufferEvent(bufferevent_get_underlying(bev), false);
}


int BufferEvent::getFD() const {return (int)bufferevent_getfd(bev);}


cb::Event::Buffer BufferEvent::getInput() const {
  return Buffer(bufferevent_get_input(bev), false);
}


cb::Event::Buffer BufferEvent::getOutput() const {
  return Buffer(bufferevent_get_output(bev), false);
}
